/*
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:      CCM (obtainGeneralisedExpectations.C)
  
  Copyright: Damian J. J. Farnell

  Date:      Tuesday May 6 2008

  Purpose:  This is a new routine that allows us to 
  calculate expectation values once the bra and the
  ket states have been solved in solve.C. Previously,
  we determined only simple values such as the sublattice
  magnetisation and differentials of the gs energy.
  
  This new code allows us to determine new equations for
  any of the terms [0,64]. This means that we can find
  s^z s^z (r), s.s (r), and even s^z or s^x etc.
  
  Crucially, the distance "r" for correlation functions
  doesn't have to be a nearest-neighbour. Thus we can
  investigate spin-spin correlations functions using 
  high-order techniques. However, for LSUBm r is approximately
  equal to 2m. Above this such correlation functions 
  ought to be constant...?
  
  This routine determines CCM equations for expectation values
  for ther data structure expectation_type/expectation.

  ===================================================

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

extern int subsetClusters(int (*config)[5], int length); // 1 if ok, 0 if not
extern void collect_term(int indices[6],double dummy,int id);

// Joerg Jun08: replaced
// int radiusGenExp=0; // only used within obtainGeneralisedExpectations.C
extern int radius;
extern int h_id;    // set to 2 here for expects, used by collect_term

// defined in collect.C
extern unsigned long
 nterms_in1,  nterms_in2,
 nterms_out1, nterms_out2;

// see at bottom for obtain(void)

//  split configuration conf0 into conf1 and conf2 according to split-array
//  take sites from conf0[1..len0] and append (!) to conf1[len1..] + conf2
extern
int  split_config(int (*conf0)[5], int  len0, int *split0,
                  int (*conf1)[5], int *len1,
                  int (*conf2)[5], int *len2);

// Jun08: Hamiltonian[bonds][term] is equivalent to expectation[expect][term]

/* test if site s1 and site s2 are neighbours k -> m */
extern
int test_k_to_m_neighbour(struct hamiltonian_defn hh[],int term, int s1[5], int s2[5]);

// termXX() replaced by obtain.C
#include "obtain.h"

extern int flag_subset; // Apr08 JS for SpeedUp
  
// same as obtain() ToDo: replace by one unique function!
void obtainGenExp(void)
{

  int    i,j,expect;
  int    length;
  int    (*configuration)[5];
  int    dummy;
  double energy;
  double lhs, rhs;
  time_t t1, t2;
  struct hamiltonian_defn *hh;  /* [MAX_TERMS62] */
  
//  static int n_out=0;   /* number of outputs */
#define MAX_out 10
  
  //  term_open(2,"w+"); Jun08

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  flag_subset=1;  // do subset-Check before findDescriptorOfCluster Apr08 JS

  if (!myid) printf("\n");

  // The data structures binary_max[], trimary_max[] ... are used 
  // to partition a given cluster into two (e.g. F(k) F(m)), three 
  // .... pieces in each `term' in H. 
  
  binary_max[0]=1;     dummy=2;   // 2^0 .. 2^30  # 30bit
  for(i=1;i<=30;i++) { binary_max[i]=dummy;   dummy=2*dummy; }

  trimary_max[0]=1;    dummy=3;   // 3^0 .. 3^10  # 16bit, i<=MAX_LSUBM
  for(i=1;i<=10;i++) { trimary_max[i]=dummy;  dummy=3*dummy; }

  quartary_max[0]=1;   dummy=4;   // 4^0 .. 4^10  # 20bit
  for(i=1;i<=10;i++) { quartary_max[i]=dummy; dummy=4*dummy; }

 // strcpy(output, Control.name);  strcat(output,".terms"); /* terms */
 // strcpy(gs_name,Control.name);  strcat(gs_name,"_gs");   /* gs terms */

  // terms_fp=fopen(output,"w"); => done in ccm.C now

  // ground-state energy equation.

  // fp_gs=fopen(gs_name,"w");  /* same content for all threads */


  /* radius need to be updated to include to longer range correlations. 
  Otherwise, the CCM equations for the expectation values won't be determined
  correctly. */
  
  // Joerg Jun08: radiusGenExp to radius (unique)
  radius=0;
  {
    int length;
    int (*tempConfig)[5]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
    for(i=1;i<=NumberConfigurations;i++) {
      length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
                                       tempConfig);
      /* note that midPointX, midPointY, midPointZ indicate Unit cells */
      // note that integer rounding shouldn't affect this 
      // algorithm
      // now find the maximum `radius'
      for(j=1;j<=length;j++) {
        if (abs(tempConfig[j][1]) > radius) radius=abs(tempConfig[j][1]);
        if (abs(tempConfig[j][2]) > radius) radius=abs(tempConfig[j][2]);
        if (abs(tempConfig[j][3]) > radius) radius=abs(tempConfig[j][3]);
      }
    }
    // radius=radius*3+1;  // ??? Joerg Jun08: to big, to slow
    radius++; // Joerg Jun08: faster and better?
    if (!myid)
      printf("\nmaximum cell radius of all clusters = %d\n\n",radius);
    free(tempConfig);
  }

  // find CCM equations.

  if (!myid){
    printf(" Write: %s.exp_terms (buffered,non-parallel)\n\n",Control.name); // Jun08
    printf("==========================================================\n\n");
    printf("OBTAIN CCM EXPECTATION EQUATIONS: START ALGORITHM.\n\n");
    printf("==========================================================\n\n");
  }

  lhs=0.0;rhs=0.0;energy=0.0;
  /* delta=1.0;  Joerg Sep05: delta seems to be not used here ??? */
  t1=time(NULL);
  h_id=2;  // Jun08 Joerg: OK???

  for(expect=1;expect<=Params.nExpect;expect++) {
    
    if (Control.output > 0 && !myid) printf("\n");

    hh=expectation[expect];  // abbreviation, same as Hamiltonian

    // find ground-state energy
    energy += /* delta * */ gse_term1(hh, 1); // term13  1 
    energy += /* delta * */ gse_term2(hh,27); // term13  9
    energy += /* delta * */ gse_term3(hh,30); // term13  9
    energy += /* delta * */ gse_term4(hh, 8); // term13  3
    energy += /* delta * */ gse_term5(hh,17); // term13  5
    energy += /* delta * */ gse_term6(hh,54); // term13  7
    energy += /* delta * */ gse_term7(hh,55); // term13 10
    energy += /* delta * */ gse_term8(hh,58); // term13 11

      int indices[7]={0};
      indices[0]=STOP_CFG;indices[1]=STOP_TERM;  /* mark end of bond data */
      indices[2]=STOP_CFG;indices[3]=STOP_CFG;
      indices[4]=STOP_CFG;indices[5]=STOP_CFG;
      collect_term(indices,0.0,2);

    for(i=1;i<=NumberConfigurations;i++) { // total number of clusters
        
	length = findClusterFromDescriptor(spinConfigurations[i].descriptor,
		   			   configuration);        
		
        hh=expectation[expect];  // abbreviation, same as Hamiltonian

	// s^z s^z : s all s = t1
	
	lhs += term2(i,length,configuration,hh,2);  // map62[2]=1
	lhs += term3(i,length,configuration,hh,3);
	lhs += term4(i,length,configuration,hh,4);
	lhs += term5(i,length,configuration,hh,5);
	
	// s^z s^+ all s = t2
	
	rhs += term6(i,length,configuration,hh,6);  // map62[6]=2
	rhs += term7(i,length,configuration,hh,7);
      
	// s^z s^- all s = t3
	
	rhs += term8(i,length,configuration,hh,8); // map62[8]=3
	rhs += term9(i,length,configuration,hh,9);
	rhs += term10(i,length,configuration,hh,10);
	rhs += term11(i,length,configuration,hh,11);
	rhs += term12(i,length,configuration,hh,12);
	rhs += term13(i,length,configuration,hh,13);
	
	// s^+ s^z all s = t4
	
	rhs += term14(i,length,configuration,hh,14); // map62[14]=4
	rhs += term15(i,length,configuration,hh,15);
      
      // s^- s^z all s
      
	rhs += term16(i,length,configuration,hh,16);
	rhs += term17(i,length,configuration,hh,17);
	rhs += term18(i,length,configuration,hh,18);
	rhs += term19(i,length,configuration,hh,19);
	rhs += term20(i,length,configuration,hh,20);
	rhs += term21(i,length,configuration,hh,21);

      // s^+ s^- all s

	rhs += term22(i,length,configuration,hh,22);
	rhs += term23(i,length,configuration,hh,23);

      // s^- s^+ all s

	rhs += term24(i,length,configuration,hh,24);
	rhs += term25(i,length,configuration,hh,25);

      
      // s^+ s^+ all s = t8

	rhs += term26(i,length,configuration,hh,26); // map62[26]=8
      
      // s^- s^- all s

	rhs += term27(i,length,configuration,hh,27); /* => length+1 */
	rhs += term28(i,length,configuration,hh,28);
	rhs += term29(i,length,configuration,hh,29);
	rhs += term30(i,length,configuration,hh,30); /* slowest 75% square */
	rhs += term31(i,length,configuration,hh,31); /* t30..34 90% of slowness */
	rhs += term32(i,length,configuration,hh,32);
	rhs += term33(i,length,configuration,hh,33);
	rhs += term34(i,length,configuration,hh,34); /* slowest 31% chain */
	rhs += term35(i,length,configuration,hh,35);

      // s^- s^- s > 1/2  = t9
      
	rhs += term36(i,length,configuration,hh,36); // map62[36]=9
	rhs += term37(i,length,configuration,hh,37);
	rhs += term38(i,length,configuration,hh,38);
	rhs += term39(i,length,configuration,hh,39);
	rhs += term40(i,length,configuration,hh,40);
	rhs += term41(i,length,configuration,hh,41);
	rhs += term42(i,length,configuration,hh,42);
	rhs += term43(i,length,configuration,hh,43);
	rhs += term44(i,length,configuration,hh,44);
	rhs += term45(i,length,configuration,hh,45);
      
      // s^z s^- s > 1/2  = t3

	rhs += term46(i,length,configuration,hh,46); // map62[46]=3
	rhs += term47(i,length,configuration,hh,47);
	rhs += term48(i,length,configuration,hh,48);

      // s^- s^z s > 1/2  = t5

	rhs += term49(i,length,configuration,hh,49); // map62[49]=5
	rhs += term50(i,length,configuration,hh,50);
	rhs += term51(i,length,configuration,hh,51);

      // s^+ s^- s > 1/2 = t6
     
	rhs += term52(i,length,configuration,hh,52); // map62[52]=6

      // s^- s^+ s > 1/2 = t7
     
	rhs += term53(i,length,configuration,hh,53); // map62[53]=7

      // s^z all s = t10

	rhs += term54(i,length,configuration,hh,54); // map62=10
     
      // s^- all s = t11

	rhs += term55(i,length,configuration,hh,55); // map62=11
	rhs += term56(i,length,configuration,hh,56);
	rhs += term57(i,length,configuration,hh,57);

      // (s^z)^2 = t12

	rhs += term58(i,length,configuration,hh,58); // map62=12
	rhs += term59(i,length,configuration,hh,59);
	rhs += term60(i,length,configuration,hh,60);
	rhs += term61(i,length,configuration,hh,61);

      // s^+     = t13

	rhs += term62(i,length,configuration,hh,62); // map62=13
	}
    
      indices[0]=STOP_CFG;indices[1]=STOP_TERM;  
      indices[2]=STOP_CFG;indices[3]=STOP_CFG;
      indices[4]=STOP_CFG;indices[5]=STOP_CFG;
      collect_term(indices,0.0,2);
   }
   
   if (Control.output>0 && myid==0)
      printf("END H[expect=%d] ==================================\n\n",expect);

   t2=time(NULL)-t1;
   if (myid==0) {
    printf("\nOBTAIN CCM EXPECTATION EQUATIONS: SUCCESSFULLY COMPLETED!\n\n");
    printf(" obtainGenExp.t/min=%.2f (jobid=0)", t2/60.);
   }

  // fclose(fp_exp); Jun08 ToDo?
  free(configuration); // memleak until v0606

}
